/*
 * test_fragment.c
 *
 * Copyright 2022 Jonathan Schöbel <jonathan@Ubermos-2019>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301, USA.
 *
 *
 */


#include <check.h>
#include <stdbool.h>
#include <stdlib.h>

#include "macro.h"
#include "error.h"
#include "fragment.h"


START_TEST(test_fragment)
{
	struct SH_Error error;
	struct SH_Fragment * fragment;
	struct SH_Data * data;
	const char * tag = "tag";
	const char * no_tag = "no_tag";

	data = SH_Data_new (NULL);
	SH_Validator_register_tag (data->validator, tag, NULL);

	/* valid tag - no error */
	fragment = SH_Fragment_new (tag, data, NULL);
	ck_assert_int_ne ((long int) fragment, (long int) NULL);

	ck_assert_str_eq (fragment->tag, tag);

	SH_Fragment_free (fragment, NULL);

	/* invalid tag - no error */
	fragment = SH_Fragment_new (no_tag, data, NULL);
	ck_assert_int_eq ((long int) fragment, (long int) NULL);

	/* valid tag - error */
	error.type = UNDEFINED;
	fragment = SH_Fragment_new (tag, data, &error);
	ck_assert_int_ne ((long int) fragment, (long int) NULL);
	ck_assert_int_eq (error.type, SUCCESS);

	ck_assert_str_eq (fragment->tag, tag);

	error.type = UNDEFINED;
	SH_Fragment_free (fragment, &error);
	ck_assert_int_eq (error.type, SUCCESS);

	/* invalid tag - error */
	error.type = UNDEFINED;
	fragment = SH_Fragment_new (no_tag, data, &error);
	ck_assert_int_eq ((long int) fragment, (long int) NULL);
	ck_assert_int_eq (error.type, VALUE_ERROR);

	SH_Data_free (data, NULL);
}
END_TEST

START_TEST(test_fragment_copy)
{
	struct SH_Error error;
	struct SH_Fragment * fragment;
	struct SH_Fragment * copy;
	struct SH_Data * data;

	data = SH_Data_new (NULL);

	fragment = SH_Fragment_new ("html", data, NULL);

	/* without error */
	copy = SH_Fragment_copy (fragment, NULL);

	ck_assert_ptr_ne (copy, NULL);
	ck_assert_ptr_ne (fragment, copy);
	ck_assert_str_eq (fragment->tag, copy->tag);
	ck_assert_int_eq (copy->child_n, 0);
	ck_assert_ptr_ne (fragment->childs, copy->childs);

	SH_Fragment_free (copy, NULL);

	/* with error */
	error.type = UNDEFINED;
	copy = SH_Fragment_copy (fragment, &error);
	ck_assert_int_eq (error.type, SUCCESS);

	ck_assert_ptr_ne (copy, NULL);
	ck_assert_ptr_ne (fragment, copy);
	ck_assert_str_eq (fragment->tag, copy->tag);
	ck_assert_int_eq (copy->child_n, 0);
	ck_assert_ptr_ne (fragment->childs, copy->childs);

	SH_Fragment_free (fragment, NULL);
	SH_Fragment_free (copy, NULL);

	SH_Data_free (data, NULL);
}
END_TEST

static void
check_childs (struct SH_Fragment * fragment, struct SH_Fragment * copy)
{
	size_t index;

	for (index = 0; index < copy->child_n; index++)
	{
		ck_assert_ptr_ne (fragment->childs[index],
				  copy->childs[index]);
		ck_assert_str_eq (fragment->childs[index]->tag,
				  copy->childs[index]->tag);
		ck_assert_int_eq (fragment->childs[index]->child_n,
				  copy->childs[index]->child_n);

		check_childs (fragment->childs[index],
			      copy->childs[index]);
	}

	return;
}

START_TEST(test_fragment_deepcopy)
{
	struct SH_Error error;
	struct SH_Fragment * fragment;
	struct SH_Fragment * child1;
	struct SH_Fragment * child2;
	struct SH_Fragment * copy;
	struct SH_Data * data;

	data = SH_Data_new (NULL);

	fragment = SH_Fragment_new ("html", data, NULL);

	child1 = SH_Fragment_new ("head", data, NULL);
	SH_Fragment_append_child (fragment, child1, NULL);
	child2 = SH_Fragment_new ("title", data, NULL);
	SH_Fragment_append_child (child1, child2, NULL);

	child1 = SH_Fragment_new ("body", data, NULL);
	SH_Fragment_append_child (fragment, child1, NULL);
	child2 = SH_Fragment_new ("header", data, NULL);
	SH_Fragment_append_child (child1, child2, NULL);
	child2 = SH_Fragment_new ("main", data, NULL);
	SH_Fragment_append_child (child1, child2, NULL);
	child2 = SH_Fragment_new ("footer", data, NULL);
	SH_Fragment_append_child (child1, child2, NULL);

	/* without error */
	copy = SH_Fragment_deepcopy (fragment, NULL);

	ck_assert_ptr_ne (copy, NULL);
	ck_assert_ptr_ne (fragment, copy);
	ck_assert_str_eq (fragment->tag, copy->tag);
	ck_assert_int_eq (fragment->child_n, copy->child_n);
	ck_assert_ptr_ne (fragment->childs, copy->childs);

	check_childs (fragment, copy);

	SH_Fragment_free (copy, NULL);

	/* with error */
	error.type = UNDEFINED;
	copy = SH_Fragment_deepcopy (fragment, &error);
	ck_assert_int_eq (error.type, SUCCESS);

	ck_assert_ptr_ne (copy, NULL);
	ck_assert_ptr_ne (fragment, copy);
	ck_assert_str_eq (fragment->tag, copy->tag);
	ck_assert_int_eq (fragment->child_n, copy->child_n);
	ck_assert_ptr_ne (fragment->childs, copy->childs);

	check_childs (fragment, copy);

	SH_Fragment_free (fragment, NULL);
	SH_Fragment_free (copy, NULL);

	SH_Data_free (data, NULL);
}
END_TEST

START_TEST(test_fragment_tag)
{
	struct SH_Error error;
	struct SH_Fragment * fragment;
	struct SH_Data * data;
	char * tag;
	const char * tag1 = "html";
	const char * tag2 = "body";

	data = SH_Data_new (NULL);
	SH_Validator_register_tag (data->validator, tag1, NULL);
	SH_Validator_register_tag (data->validator, tag2, NULL);

	/* no error */
	fragment = SH_Fragment_new (tag1, data, NULL);

	tag = SH_Fragment_get_tag (fragment, NULL);
	ck_assert_str_eq (tag, tag1);
	free (tag);

	SH_Fragment_free (fragment, NULL);

	/* error */
	fragment = SH_Fragment_new (tag2, data, NULL);

	error.type = UNDEFINED;
	tag = SH_Fragment_get_tag (fragment, &error);
	ck_assert_str_eq (tag, tag2);
	ck_assert_int_eq (error.type, SUCCESS);
	free (tag);

	SH_Fragment_free (fragment, NULL);
	SH_Data_free (data, NULL);
}
END_TEST

START_TEST(test_fragment_child)
{
	struct SH_Error error;
	struct SH_Fragment * parent;
	struct SH_Fragment * child1;
	struct SH_Fragment * child2;
	struct SH_Data * data;
	bool boolean;

	data = SH_Data_new (NULL);

	parent = SH_Fragment_new ("html", data, NULL);
	child1 = SH_Fragment_new ("head", data, NULL);
	child2 = SH_Fragment_new ("body", data, NULL);

	/* no error */
	ck_assert_int_eq (parent->child_n, 0);

	boolean = SH_Fragment_append_child (parent, child1, NULL);
	ck_assert_int_eq (boolean, TRUE);
	ck_assert_int_eq (parent->child_n, 1);
	ck_assert_ptr_eq (parent->childs[0], child1);

	/* with error */
	ck_assert_int_eq (parent->child_n, 1);

	error.type = UNDEFINED;
	boolean = SH_Fragment_append_child (parent, child2, &error);
	ck_assert_int_eq (boolean, TRUE);
	ck_assert_int_eq (error.type, SUCCESS);
	ck_assert_int_eq (parent->child_n, 2);
	ck_assert_ptr_eq (parent->childs[1], child2);

	SH_Fragment_free (parent, NULL);

	SH_Data_free (data, NULL);
}
END_TEST

START_TEST(test_fragment_get_child)
{
	struct SH_Error error;
	struct SH_Fragment * parent;
	struct SH_Fragment * child1;
	struct SH_Fragment * child2;
	struct SH_Data * data;

	data = SH_Data_new (NULL);

	parent = SH_Fragment_new ("html", data, NULL);
	child1 = SH_Fragment_new ("body", data, NULL);

	SH_Fragment_append_child (parent, child1, NULL);

	/* without error */
	child2 = SH_Fragment_get_child (parent, 0, NULL);
	ck_assert_ptr_eq (child1, child2);

	child2 = SH_Fragment_get_child (parent, 1, NULL);
	ck_assert_ptr_eq (NULL, child2);

	/* with error */
	error.type = UNDEFINED;
	child2 = SH_Fragment_get_child (parent, 0, &error);
	ck_assert_ptr_eq (child1, child2);
	ck_assert_int_eq (error.type, SUCCESS);

	error.type = UNDEFINED;
	child2 = SH_Fragment_get_child (parent, 1, &error);
	ck_assert_ptr_eq (NULL, child2);
	ck_assert_int_eq (error.type, VALUE_ERROR);

	SH_Fragment_free (parent, NULL);

	SH_Data_free (data, NULL);
}
END_TEST

START_TEST(test_fragment_is_child)
{
	struct SH_Fragment * parent;
	struct SH_Fragment * child1;
	struct SH_Fragment * child2;
	struct SH_Data * data;
	bool boolean;

	data = SH_Data_new (NULL);

	parent = SH_Fragment_new ("html", data, NULL);
	child1 = SH_Fragment_new ("head", data, NULL);
	child2 = SH_Fragment_new ("title", data, NULL);

	SH_Fragment_append_child (parent, child1, NULL);
	SH_Fragment_append_child (child1, child2, NULL);

	boolean = SH_Fragment_is_child (parent, child1);
	ck_assert_int_eq (boolean, TRUE);

	boolean = SH_Fragment_is_child (parent, child2);
	ck_assert_int_eq (boolean, FALSE);

	boolean = SH_Fragment_is_child (child1, child2);
	ck_assert_int_eq (boolean, TRUE);

	SH_Fragment_free (parent, NULL);

	SH_Data_free (data, NULL);
}
END_TEST

START_TEST(test_fragment_is_descendant)
{
	struct SH_Fragment * parent;
	struct SH_Fragment * child1;
	struct SH_Fragment * child2;
	struct SH_Fragment * child3;
	struct SH_Data * data;
	bool boolean;

	data = SH_Data_new (NULL);

	parent = SH_Fragment_new ("html", data, NULL);
	child1 = SH_Fragment_new ("head", data, NULL);
	child2 = SH_Fragment_new ("body", data, NULL);
	child3 = SH_Fragment_new ("title", data, NULL);

	SH_Fragment_append_child (parent, child1, NULL);
	SH_Fragment_append_child (parent, child2, NULL);
	SH_Fragment_append_child (child1, child3, NULL);

	boolean = SH_Fragment_is_descendant (parent, child1);
	ck_assert_int_eq (boolean, TRUE);

	boolean = SH_Fragment_is_descendant (parent, child2);
	ck_assert_int_eq (boolean, TRUE);

	boolean = SH_Fragment_is_descendant (parent, child3);
	ck_assert_int_eq (boolean, TRUE);

	boolean = SH_Fragment_is_descendant (child1, child2);
	ck_assert_int_eq (boolean, FALSE);

	boolean = SH_Fragment_is_descendant (child1, child3);
	ck_assert_int_eq (boolean, TRUE);

	boolean = SH_Fragment_is_descendant (child2, child3);
	ck_assert_int_eq (boolean, FALSE);

	SH_Fragment_free (parent, NULL);

	SH_Data_free (data, NULL);
}
END_TEST

START_TEST(test_fragment_html)
{
	struct SH_Error error;
	struct SH_Fragment * fragment1;
	struct SH_Fragment * fragment2;
	struct SH_Data * data;
	struct SH_Text * text;

	data = SH_Data_new (NULL);
	SH_Validator_register_tag (data->validator, "html", NULL);

	/* no error */
	fragment1 = SH_Fragment_new ("html", data, NULL);
	fragment2 = SH_Fragment_new ("body", data, NULL);
	SH_Fragment_append_child (fragment1, fragment2, NULL);

	text = SH_Fragment_to_html (fragment1, INLINE, NULL);
	ck_assert_str_eq (text->text, "<html><body></body></html>");

	text = SH_Fragment_to_html (fragment1, WRAP, NULL);
	ck_assert_str_eq (text->text,
			  "<html>\n\t<body>\n\t</body>\n</html>\n");

	SH_Fragment_free (fragment1, NULL);

	/* error */
	fragment1 = SH_Fragment_new ("html", data, NULL);
	fragment2 = SH_Fragment_new ("body", data, NULL);
	SH_Fragment_append_child (fragment1, fragment2, NULL);

	error.type = UNDEFINED;
	text = SH_Fragment_to_html (fragment1, INLINE, &error);
	ck_assert_str_eq (text->text, "<html><body></body></html>");
	ck_assert_int_eq (error.type, SUCCESS);

	error.type = UNDEFINED;
	text = SH_Fragment_to_html (fragment1, WRAP, &error);
	ck_assert_str_eq (text->text,
			  "<html>\n\t<body>\n\t</body>\n</html>\n");
	ck_assert_int_eq (error.type, SUCCESS);

	SH_Fragment_free (fragment1, NULL);
	SH_Data_free (data, NULL);
}
END_TEST

Suite * fragment_suite (void)
{
	Suite *s;
	TCase *tc_core;

	s = suite_create ("Testsuite SeFHT Fragment");

	/* Core test case */
	tc_core = tcase_create ("Core");

	tcase_add_test (tc_core, test_fragment);
	tcase_add_test (tc_core, test_fragment_copy);
	tcase_add_test (tc_core, test_fragment_deepcopy);
	tcase_add_test (tc_core, test_fragment_tag);
	tcase_add_test (tc_core, test_fragment_child);
	tcase_add_test (tc_core, test_fragment_get_child);
	tcase_add_test (tc_core, test_fragment_is_child);
	tcase_add_test (tc_core, test_fragment_is_descendant);
	tcase_add_test (tc_core, test_fragment_html);
	suite_add_tcase (s, tc_core);

	return s;
}

int main (int argc, char **argv)
{
	int number_failed;
	Suite *s;
	SRunner *sr;

	s = fragment_suite ();
	sr = srunner_create (s);

	srunner_run_all (sr, CK_NORMAL);
	number_failed = srunner_ntests_failed (sr);
	srunner_free (sr);

	return (number_failed == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}

